# -*- coding:utf-8 -*-
# pylint: disable=line-too-long

_BASE_REPRESENTATIONS = [
    "Primitive(field='f1', op='eq', value='value')",
    "Primitive(field='f1', op='eq', value=42)",
    "Primitive(field='f1', op='eq', value=3.14)",
    "Primitive(field='f1', op='eq', value=True)",
    "Condition(op=Operator.OR, values=[Primitive(field='f1', op='eq', value='aaa'), Primitive(field='f2', op='eq', value=5)])",
    "Condition(op=Operator.OR, values=[Primitive(field='f1', op='eq', value='aaa'), Primitive(field='f2', op='eq', value=5), Primitive(field='f1', op='eq', value='bbb')])",
    "Condition(op=Operator.AND, values=[Primitive(field='f1', op='eq', value='aaa'), Primitive(field='f2', op='eq', value=5)])",
    "Condition(op=Operator.AND, values=[Primitive(field='f1', op='eq', value='aaa'), Primitive(field='f2', op='eq', value=5), Primitive(field='f1', op='eq', value='bbb')])",
    "Condition(op=Operator.OR, values=[Condition(op=Operator.AND, values=[Primitive(field='f1', op='eq', value=50), Primitive(field='f2', op='eq', value='aaa')]), Primitive(field='f2', op='eq', value='bbb')])",
]

REPRESENTATIONS = _BASE_REPRESENTATIONS + [
    val.replace(
        "field='f1', op='eq", "field='f1', op='gt"
    ).replace(
        "field='f2', op='eq'", "field='f2', op='match'"
    )
    for val in _BASE_REPRESENTATIONS
]



# -*- coding:utf-8 -*-
from copy import deepcopy
from itertools import combinations_with_replacement

import pytest

from anywhere import testsets


@pytest.mark.parametrize(
    'where_obj,string',
    zip(
        testsets.WHERES,
        testsets.STRINGS,
    ),
    ids=[str(i) for i in range(1, len(testsets.WHERES) + 1)],
)
def test_str(where_obj, string):
    assert str(where_obj) == string


@pytest.mark.parametrize(
    'where_obj,representation',
    zip(
        testsets.WHERES,
        testsets.REPRESENTATIONS,
    ),
    ids=[str(i) for i in range(1, len(testsets.WHERES) + 1)],
)
def test_repr(where_obj, representation):
    assert repr(where_obj) == representation


@pytest.mark.parametrize(
    'where_obj1,where_obj2',
    combinations_with_replacement(testsets.WHERES, 2),
    ids=lambda _: '',
)
def test_comparision(where_obj1, where_obj2):
    if id(where_obj1) == id(where_obj2):
        # We are comparing the same object, should be equal
        assert where_obj1 == where_obj2
    else:
        assert where_obj1 != where_obj2


@pytest.mark.parametrize(
    'where_obj',
    testsets.WHERES,
    ids=[str(i) for i in range(1, len(testsets.WHERES) + 1)],
)
def test_comparision_deepcopy(where_obj):
    where_obj2 = deepcopy(where_obj)

    assert id(where_obj) != id(where_obj2)
    assert where_obj == where_obj2



# -*- coding:utf-8 -*-
"""
Protocol buffers utils.
"""
import typing

from anywhere import base_types as bt
from . import anywhere_pb2 as pb


def serialize(where: typing.Union[bt.Primitive, bt.Condition]) -> pb.Where:
    """

    :param where:
    :return:
    """
    # pylint: disable=no-member
    msg = pb.Where()

    if isinstance(where, bt.Primitive):
        msg.primitive_value.field = where.field
        msg.primitive_value.op = where.op

        if isinstance(where.value, bool):
            msg.primitive_value.bool_value = where.value
        elif isinstance(where.value, int):
            msg.primitive_value.int_value = where.value
        elif isinstance(where.value, float):
            msg.primitive_value.double_value = where.value
        else:
            msg.primitive_value.string_value = where.value
    else:
        msg.condition_value.op = pb.Operator.AND if where.op == bt.Operator.AND else pb.Operator.OR
        msg.condition_value.values.extend([
            serialize(v) for v in where.values
        ])

    return msg


def load(msg: pb.Where) -> typing.Union[bt.Primitive, bt.Condition]:
    """

    :param msg:
    :return:
    """
    if msg.HasField('primitive_value'):
        val = msg.primitive_value

        value_variant = val.WhichOneof('value')
        where = bt.Primitive(msg.primitive_value.field, msg.primitive_value.op, getattr(val, value_variant))
    else:
        val = msg.condition_value

        if val.op == pb.Operator.AND:
            operator = bt.Operator.AND
        else:
            operator = bt.Operator.OR

        where = bt.Condition(operator, [load(v) for v in val.values])

    return where



# -*- coding:utf-8 -*-
import typing

from sqlalchemy import select, Table, and_, or_

from anywhere import base_types as bt
from .field import Field


def _filter_fields(attrs):
    """
    """
    return filter(lambda item: isinstance(item[1], Field), attrs.items())


def _frozen_setattrs(self, name, value):
    raise RuntimeError('Meta if already frozen')


class QueryProducerMeta(type):

    def __new__(mcs, name, bases, attrs: typing.MutableMapping):  # pylint: disable=bad-mcs-classmethod-argument
        print('bases:', bases)

        if name == 'QueryProducer':
            return super().__new__(mcs, name, bases, attrs)

        attrs['_meta'] = mcs.prepare_meta(attrs)
        attrs['_fields'] = mcs.prepare_fields_map(attrs)

        return super().__new__(mcs, name, bases, attrs)

    @staticmethod
    def prepare_meta(attrs: typing.MutableMapping):
        # pylint: disable=unsupported-membership-test,unsubscriptable-object
        if 'Meta' not in attrs:
            raise RuntimeError('no Meta')
        if not hasattr(attrs['Meta'], 'table'):
            raise RuntimeError('no table in Meta')
        if not isinstance(attrs['Meta'].table, Table):
            raise RuntimeError('Meta.table is not Table')

        frozen_meta = attrs.pop('Meta')
        frozen_meta.__setattr__ = _frozen_setattrs
        frozen_meta.__delattr__ = _frozen_setattrs

        return frozen_meta

    @staticmethod
    def prepare_fields_map(attrs: typing.MutableMapping):
        # pylint: disable=unsubscriptable-object
        table = attrs['_meta'].table
        fields = {}

        for name, field in _filter_fields(attrs):
            field.name = name
            fields[name] = field

            if field.column is not None:
                continue

            if name not in table.columns:
                raise RuntimeError(f'Column {name} does not exist in {table}!')

            field.column = table.columns[name]

        return fields


class MetaType(typing.NamedTuple):
    table: Table


class QueryProducer(metaclass=QueryProducerMeta):

    _fields: typing.Mapping[str, Field] = None
    _meta: MetaType

    # Public interface

    @property
    def fields(self) -> typing.Mapping[str, Field]:
        return self._fields

    def get_base_query(self):
        return select([self._meta.table])

    def query_from_where(self, where, *, base_query=None):
        if base_query is None:
            query = self.get_base_query()
        else:
            query = base_query

        return query.where(self._whereclause(where))

    # Internal functions

    def _whereclause(self, where):
        if isinstance(where, bt.Primitive):
            return self._primitive_comparision(where)

        return self._logical_condition(where)

    def _primitive_comparision(self, where: bt.Primitive):
        field = self.fields[where.field]  # pylint: disable=unsubscriptable-object
        cmp_fn = field.operators[where.op]

        return cmp_fn(field.column, where.value)

    def _logical_condition(self, where: bt.Condition):
        operator = and_ if where.op == bt.Operator.AND else or_

        return (operator(*[
            self._whereclause(child) for child in where.values
        ]))



# -*- coding:utf-8 -*-
import operator

DEFAULT_OPERATORS = (
    ('eq', operator.eq),
    ('neq', operator.ne),
    ('gt', operator.gt),
    ('gte', operator.ge),
    ('lt', operator.lt),
    ('lte', operator.le),
)


class Field:
    # pylint: disable=too-few-public-methods

    _name: str = None

    def __init__(self, *, column=None, operators=DEFAULT_OPERATORS, extra_operators=None):
        ops = dict(operators)
        if extra_operators:
            ops.update(extra_operators)

        self.column = column
        self.operators = ops

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if self._name is not None:
            raise RuntimeError('name already set')
        self._name = value



# -*- coding:utf-8 -*-
import pytest

from anywhere import testsets
from . import testset as local_testset


@pytest.mark.parametrize(
    'where_obj,expected_query',
    zip(
        testsets.WHERES,
        local_testset.QUERIES,
    ),
    ids=[str(i) for i in range(1, len(testsets.WHERES) + 1)],
)
def test_produce_query(where_obj, expected_query, query_producer):
    parsed = query_producer.query_from_where(where_obj)

    assert str(parsed) == str(expected_query)



# -*- coding:utf-8 -*-
"""
Classes serving as a common representation of where queries.

All new formats or third-party libraries support should be implemented
by working with these classes (Operator, Primitive, Condition).

Third-party implementation means:
 1. Serializing them to desired format (protobuf binary, SQL query).
 2. Loading them from desired format (protobuf binary, graphql query).
"""
import enum
import typing
from dataclasses import dataclass


class Operator(enum.Enum):
    """
    Logical operators available in where queries.
    """
    AND = 'AND'
    OR = 'OR'

    def __repr__(self):
        return f'Operator.{self.value}'


@dataclass
class Primitive:
    """
    Primitive is a leaf value in grapqhql where query argument. It consists of field name, operator, and value.

    For example:

    ```
    { title: "Pinokio" }
    ```

    `eq` operator is used by default. Different operators might are used by adding postfix to field name:

    ```
    { title__neq: "Pinokio" }
    ```
    """
    field: str
    op: str  # pylint: disable=invalid-name
    value: typing.Union[str, float, int, bool]

    def __repr__(self):
        return f'Primitive(field={repr(self.field)}, op={repr(self.op)}, value={repr(self.value)})'

    def __str__(self):
        return f'{self.field} {self.op} {repr(self.value)}'

    def __eq__(self, other):
        if not isinstance(other, Primitive):
            return NotImplemented

        return (
            self.field == other.field and
            self.op == other.op and
            self.value == other.value
        )


@dataclass
class Condition:
    """
    Logical conjunction of Primitives or other Operators.

    For example:

    ```
    { OR: [{ title: "Pinokio"}, { title: "Little Prince"}] }
    ```
    """
    op: Operator  # pylint: disable=invalid-name
    values: typing.Iterable[typing.Union[Primitive, 'Condition']]

    def __repr__(self):
        return f'Condition(op={repr(self.op)}, values={repr(self.values)})'

    def __str__(self):
        operator = ' & ' if self.op == Operator.AND else ' | '

        return '(' + (operator.join([str(v) for v in self.values])) + ')'

    def __eq__(self, other):
        if not isinstance(other, Condition):
            return NotImplemented

        return self.op == other.op and self.values == other.values



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: anywhere/anywhere.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='anywhere/anywhere.proto',
  package='anywhere',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x17\x61nywhere/anywhere.proto\x12\x08\x61nywhere\"p\n\x05Where\x12.\n\x0fprimitive_value\x18\x01 \x01(\x0b\x32\x13.anywhere.PrimitiveH\x00\x12.\n\x0f\x63ondition_value\x18\x02 \x01(\x0b\x32\x13.anywhere.ConditionH\x00\x42\x07\n\x05value\"\x8a\x01\n\tPrimitive\x12\r\n\x05\x66ield\x18\x01 \x01(\t\x12\n\n\x02op\x18\x02 \x01(\t\x12\x13\n\tint_value\x18\x03 \x01(\x03H\x00\x12\x16\n\x0c\x64ouble_value\x18\x04 \x01(\x01H\x00\x12\x16\n\x0cstring_value\x18\x05 \x01(\tH\x00\x12\x14\n\nbool_value\x18\x06 \x01(\x08H\x00\x42\x07\n\x05value\"L\n\tCondition\x12\x1e\n\x02op\x18\x01 \x01(\x0e\x32\x12.anywhere.Operator\x12\x1f\n\x06values\x18\x02 \x03(\x0b\x32\x0f.anywhere.Where*\x1b\n\x08Operator\x12\x07\n\x03\x41ND\x10\x00\x12\x06\n\x02OR\x10\x01\x62\x06proto3')
)

_OPERATOR = _descriptor.EnumDescriptor(
  name='Operator',
  full_name='anywhere.Operator',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='AND', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OR', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=370,
  serialized_end=397,
)
_sym_db.RegisterEnumDescriptor(_OPERATOR)

Operator = enum_type_wrapper.EnumTypeWrapper(_OPERATOR)
AND = 0
OR = 1



_WHERE = _descriptor.Descriptor(
  name='Where',
  full_name='anywhere.Where',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='primitive_value', full_name='anywhere.Where.primitive_value', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='condition_value', full_name='anywhere.Where.condition_value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='anywhere.Where.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=37,
  serialized_end=149,
)


_PRIMITIVE = _descriptor.Descriptor(
  name='Primitive',
  full_name='anywhere.Primitive',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='field', full_name='anywhere.Primitive.field', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='op', full_name='anywhere.Primitive.op', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int_value', full_name='anywhere.Primitive.int_value', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='double_value', full_name='anywhere.Primitive.double_value', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_value', full_name='anywhere.Primitive.string_value', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bool_value', full_name='anywhere.Primitive.bool_value', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='anywhere.Primitive.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=152,
  serialized_end=290,
)


_CONDITION = _descriptor.Descriptor(
  name='Condition',
  full_name='anywhere.Condition',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='op', full_name='anywhere.Condition.op', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='values', full_name='anywhere.Condition.values', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=292,
  serialized_end=368,
)

_WHERE.fields_by_name['primitive_value'].message_type = _PRIMITIVE
_WHERE.fields_by_name['condition_value'].message_type = _CONDITION
_WHERE.oneofs_by_name['value'].fields.append(
  _WHERE.fields_by_name['primitive_value'])
_WHERE.fields_by_name['primitive_value'].containing_oneof = _WHERE.oneofs_by_name['value']
_WHERE.oneofs_by_name['value'].fields.append(
  _WHERE.fields_by_name['condition_value'])
_WHERE.fields_by_name['condition_value'].containing_oneof = _WHERE.oneofs_by_name['value']
_PRIMITIVE.oneofs_by_name['value'].fields.append(
  _PRIMITIVE.fields_by_name['int_value'])
_PRIMITIVE.fields_by_name['int_value'].containing_oneof = _PRIMITIVE.oneofs_by_name['value']
_PRIMITIVE.oneofs_by_name['value'].fields.append(
  _PRIMITIVE.fields_by_name['double_value'])
_PRIMITIVE.fields_by_name['double_value'].containing_oneof = _PRIMITIVE.oneofs_by_name['value']
_PRIMITIVE.oneofs_by_name['value'].fields.append(
  _PRIMITIVE.fields_by_name['string_value'])
_PRIMITIVE.fields_by_name['string_value'].containing_oneof = _PRIMITIVE.oneofs_by_name['value']
_PRIMITIVE.oneofs_by_name['value'].fields.append(
  _PRIMITIVE.fields_by_name['bool_value'])
_PRIMITIVE.fields_by_name['bool_value'].containing_oneof = _PRIMITIVE.oneofs_by_name['value']
_CONDITION.fields_by_name['op'].enum_type = _OPERATOR
_CONDITION.fields_by_name['values'].message_type = _WHERE
DESCRIPTOR.message_types_by_name['Where'] = _WHERE
DESCRIPTOR.message_types_by_name['Primitive'] = _PRIMITIVE
DESCRIPTOR.message_types_by_name['Condition'] = _CONDITION
DESCRIPTOR.enum_types_by_name['Operator'] = _OPERATOR
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Where = _reflection.GeneratedProtocolMessageType('Where', (_message.Message,), dict(
  DESCRIPTOR = _WHERE,
  __module__ = 'anywhere.anywhere_pb2'
  # @@protoc_insertion_point(class_scope:anywhere.Where)
  ))
_sym_db.RegisterMessage(Where)

Primitive = _reflection.GeneratedProtocolMessageType('Primitive', (_message.Message,), dict(
  DESCRIPTOR = _PRIMITIVE,
  __module__ = 'anywhere.anywhere_pb2'
  # @@protoc_insertion_point(class_scope:anywhere.Primitive)
  ))
_sym_db.RegisterMessage(Primitive)

Condition = _reflection.GeneratedProtocolMessageType('Condition', (_message.Message,), dict(
  DESCRIPTOR = _CONDITION,
  __module__ = 'anywhere.anywhere_pb2'
  # @@protoc_insertion_point(class_scope:anywhere.Condition)
  ))
_sym_db.RegisterMessage(Condition)


# @@protoc_insertion_point(module_scope)



# -*- coding:utf-8 -*-
_BASE_STRINGS = [
    "f1 eq 'value'",
    "f1 eq 42",
    "f1 eq 3.14",
    "f1 eq True",
    "(f1 eq 'aaa' | f2 eq 5)",
    "(f1 eq 'aaa' | f2 eq 5 | f1 eq 'bbb')",
    "(f1 eq 'aaa' & f2 eq 5)",
    "(f1 eq 'aaa' & f2 eq 5 & f1 eq 'bbb')",
    "((f1 eq 50 & f2 eq 'aaa') | f2 eq 'bbb')",
]

STRINGS = _BASE_STRINGS + [
    val.replace('f1 eq', 'f1 gt').replace('f2 eq', 'f2 match')
    for val in _BASE_STRINGS
]



# -*- coding:utf-8 -*-
import pytest

from anywhere.protobuf3 import load, serialize
from anywhere import testsets


@pytest.mark.parametrize(
    'where_obj',
    testsets.WHERES,
    ids=[str(i) for i in range(1, len(testsets.WHERES) + 1)],
)
def test_load_serialize(where_obj):
    where_obj2 = load(serialize(where_obj))

    assert id(where_obj) != id(where_obj2)
    assert where_obj == where_obj2



# -*- coding:utf-8 -*-
from sqlalchemy import or_, and_, select

from .conftest import SomeModel


WHERECLAUSES = [
    SomeModel.__table__.c.public_id == 'value',
    SomeModel.__table__.c.public_id == 42,
    SomeModel.__table__.c.public_id == 3.14,
    SomeModel.__table__.c.public_id == True,  # pylint: disable=singleton-comparison
    or_(
        SomeModel.__table__.c.public_id == 'aaa',
        SomeModel.__table__.c.f2 == 5,
    ),
    or_(
        SomeModel.__table__.c.public_id == 'aaa',
        SomeModel.__table__.c.f2 == 5,
        SomeModel.__table__.c.public_id == 'bbb',
    ),
    and_(
        SomeModel.__table__.c.public_id == 'aaa',
        SomeModel.__table__.c.f2 == 5,
    ),
    and_(
        SomeModel.__table__.c.public_id == 'aaa',
        SomeModel.__table__.c.f2 == 5,
        SomeModel.__table__.c.public_id == 'bbb',
    ),
    or_(
        and_(
            SomeModel.__table__.c.public_id == 50,
            SomeModel.__table__.c.f2 == 'aaa',
        ),
        SomeModel.__table__.c.f2 == 'bbb',
    ),
    # custom operators
    SomeModel.__table__.c.public_id > 'value',
    SomeModel.__table__.c.public_id > 42,
    SomeModel.__table__.c.public_id > 3.14,
    SomeModel.__table__.c.public_id > 'true',  # we cannot use > operator with boolean value
    or_(
        SomeModel.__table__.c.public_id > 'aaa',
        SomeModel.__table__.c.f2.ilike('%5%'),
    ),
    or_(
        SomeModel.__table__.c.public_id > 'aaa',
        SomeModel.__table__.c.f2.ilike('%5%'),
        SomeModel.__table__.c.public_id > 'bbb',
    ),
    and_(
        SomeModel.__table__.c.public_id > 'aaa',
        SomeModel.__table__.c.f2.ilike('%5%'),
    ),
    and_(
        SomeModel.__table__.c.public_id > 'aaa',
        SomeModel.__table__.c.f2.ilike('%5%'),
        SomeModel.__table__.c.public_id > 'bbb',
    ),
    or_(
        and_(
            SomeModel.__table__.c.public_id > 50,
            SomeModel.__table__.c.f2.ilike('%aaa%'),
        ),
        SomeModel.__table__.c.f2.ilike('%bbb%'),
    ),
]

QUERIES = [
    select([SomeModel.__table__]).where(whereclause) for whereclause in WHERECLAUSES
]



# -*- coding:utf-8 -*-
import re
from setuptools_scm import get_version


if __name__ == '__main__':
    version = get_version('..', relative_to=__file__)
    result = re.match(r'^\d+.\d+.\d+(-rc.\d)?$', version)

    print('YES' if result else 'NO')



# -*- coding:utf-8 -*-
import distutils
import os
import subprocess
import sys

from setuptools import find_packages, setup


os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


with open('README.md') as fh:
    README = fh.read()


class PylintCommand(distutils.cmd.Command):
    """A custom command to run Pylint on all Python source files."""

    description = 'Run pyint linter.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run command."""
        command = [
            'pylint', '--rcfile=toolscfg/pylintrc',
            # Those three are only cluttering the results, will be fixed some day.
            '--disable', 'missing-module-docstring',
            '--disable', 'missing-function-docstring',
            '--disable', 'missing-class-docstring',
            'anywhere',
        ]

        self.announce(f'Running command: {" ".join(command)}', level=distutils.log.INFO)

        try:
            subprocess.check_call(command)
        except subprocess.CalledProcessError as ex:
            sys.exit(ex.returncode)


class CleanCommand(distutils.cmd.Command):
    """A custom command to clean build / test artifacts."""

    description = 'Clean artifacts (reports, builds...).'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run command."""
        commands = (
            ['rm', '-Rf', '.pytest_cache'],
            ['rm', '-Rf', 'htmlcov'],
            ['rm', '-Rf', 'htmltest'],
            ['rm', '-f', '.coverage'],
            ['rm', '-f', 'coverage.xml'],
            ['rm', '-f', 'report.xml'],
            ['rm', '-Rf', 'build'],
            ['rm', '-Rf', 'dist'],
        )

        for command in commands:
            self.announce(f'Running command: {" ".join(command)}', level=distutils.log.INFO)
            try:
                subprocess.check_call(command)
            except subprocess.CalledProcessError as ex:
                sys.exit(ex.returncode)


class ProtocCommand(distutils.cmd.Command):
    """A custom command to clean build / test artifacts."""

    description = 'Clean artifacts (reports, builds...).'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run command."""
        commands = (
            ['protoc', '-I', '.', '--python_out', '.', 'anywhere/anywhere.proto'],
            ['mv', 'anywhere/anywhere_pb2.py', 'anywhere/protobuf3/'],
        )

        for command in commands:
            self.announce(f'Running command: {" ".join(command)}', level=distutils.log.INFO)
            try:
                subprocess.check_call(command)
            except subprocess.CalledProcessError as ex:
                sys.exit(ex.returncode)


setup(
    name='anywhere',
    use_scm_version=True,
    description='Turn logical sentences into serializable, executable queries.',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/mlga/anywhere',
    author='mlga',
    author_email='github@mlga.io',
    license='MIT',
    keywords='filtering sqlalchemy query graphene protobuf',
    setup_requires=[
        'pytest-runner~=5.1',
        'setuptools_scm~=3.3',
    ],
    install_requires=[],
    tests_require=[
        'pytest~=5.0',
        'pytest-cov~=2.7',
        'pytest-html~=1.20',
    ],
    extras_require={
        'sqlalchemy': [
            'SQLAlchemy~=1.3.0',
        ],
        'graphene': [
            'marshmallow~=2.19',
            'graphene~=2.1',
        ],
        'protobuf': [
            'protobuf~=3.9',
        ],
        'develop': [
            'pylint~=2.3',
            'pytest~=5.0',
            'pytest-cov~=2.7',
            'pytest-html~=1.20',
        ],
    },
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    platforms='any',
    zip_safe=False,
    cmdclass={
        'pylint': PylintCommand,
        'clean': CleanCommand,
        'protoc': ProtocCommand,
    },
    python_requires='>=3.7',
)



# -*- coding:utf-8 -*-
# pylint: disable=too-few-public-methods
import pytest
from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

from anywhere.sqlalchemy import Field
from anywhere.sqlalchemy import QueryProducer

Base = declarative_base()  # pylint: disable=invalid-name


class SomeModel(Base):
    __tablename__ = 'some_table'

    id = Column(Integer, primary_key=True)
    public_id = Column(String(50))
    f2 = Column(String(50))


class SampleQueryProducer(QueryProducer):

    class Meta:
        table = SomeModel.__table__

    f1 = Field(
        column=SomeModel.__table__.c.public_id,
        extra_operators=(
            ('gt', lambda col, val: col > str(val)),  # casting to string to handle > operator with boolean value
        ),
    )
    f2 = Field(
        extra_operators=(
            ('match', lambda col, val: col.ilike(f'%{val}%')),
        ),
    )


@pytest.fixture
def query_producer():
    return SampleQueryProducer()



# -*- coding:utf-8 -*-
from copy import deepcopy

from anywhere import base_types as bt


def replace_op(where):
    where = deepcopy(where)

    op_replacement = {
        'f1': 'gt',
        'f2': 'match',
    }
    if isinstance(where, bt.Primitive):
        where.op = op_replacement.get(where.field, 'eq')
    else:
        where.values = [
            replace_op(field) for field in where.values
        ]

    return where


_BASE_WHERES = [
    bt.Primitive(field='f1', op='eq', value='value'),
    bt.Primitive(field='f1', op='eq', value=42),
    bt.Primitive(field='f1', op='eq', value=3.14),
    bt.Primitive(field='f1', op='eq', value=True),
    bt.Condition(
        op=bt.Operator.OR,
        values=[
            bt.Primitive(field='f1', op='eq', value='aaa'),
            bt.Primitive(field='f2', op='eq', value=5)
        ]
    ),
    bt.Condition(
        op=bt.Operator.OR,
        values=[
            bt.Primitive(field='f1', op='eq', value='aaa'),
            bt.Primitive(field='f2', op='eq', value=5),
            bt.Primitive(field='f1', op='eq', value='bbb')
        ]
    ),
    bt.Condition(
        op=bt.Operator.AND,
        values=[
            bt.Primitive(field='f1', op='eq', value='aaa'),
            bt.Primitive(field='f2', op='eq', value=5)
        ]
    ),
    bt.Condition(
        op=bt.Operator.AND,
        values=[
            bt.Primitive(field='f1', op='eq', value='aaa'),
            bt.Primitive(field='f2', op='eq', value=5),
            bt.Primitive(field='f1', op='eq', value='bbb')
        ]
    ),
    bt.Condition(
        op=bt.Operator.OR,
        values=[
            bt.Condition(
                op=bt.Operator.AND,
                values=[
                    bt.Primitive(field='f1', op='eq', value=50),
                    bt.Primitive(field='f2', op='eq', value='aaa')
                ]
            ),
            bt.Primitive(field='f2', op='eq', value='bbb')
        ]
    ),
]

WHERES = _BASE_WHERES + [
    replace_op(where) for where in _BASE_WHERES
]
